package com.microsoft.mobileexperiences.bing.httpthreadpool;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.Future;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

public abstract class HttpRequest implements Runnable{

	public enum HttpMethod
	{
		HTTP_METHOD_GET, 
		HTTP_METHOD_POST
	}

	private String url;
	private HttpMethod httpMethod;
	private Boolean gzipUploadStream = false;
	private int timeOut = 30000; //30 seconds
	
	private Future<?> future = null;
	private Hashtable<String, String> headers = new Hashtable<String, String>();
	private HttpResponseBase httpResponse;
	private IHttpResult httpCallback;
	private boolean httpChunkedStreamingMode = false;
	private int m_ChunkedStreamSize = 4 * 1024;
	
	public HttpRequest(String url, HttpMethod httpMethod)
	{
		this.url = url;
		this.httpMethod = httpMethod;
	}
	
	public void setUrl(String url)
	{
		this.url = url;
	}
	
	public void setCallback(IHttpResult httpCallback)
	{
		this.httpCallback = httpCallback;
	}
	
	public void setFuture(Future<?> future) //used by thread pool class. 
	{
		this.future = future;
	}
	
	public void setChunkedStreamingMode(boolean mode)
	{
		this.httpChunkedStreamingMode = mode;
	}
	
	public void setChunkedStreamSize(int size)
	{
		if (size > 0)
		{
			m_ChunkedStreamSize = size;
			httpChunkedStreamingMode = true;
		}
		else
			httpChunkedStreamingMode = false;
	}
	
	public Future<?> getFuture()  
	{
		return future;
	}
	
	public HttpResponseBase getResponse()
	{
		return httpResponse;
	}

	//default decode function. Could be overrided.
	protected HttpResponseBase decode(InputStream stream, Hashtable<String, String> headers)
	{
		HttpDefaultResponse response = new HttpDefaultResponse(stream, headers);
		return response;
	}

	public String encode() {return null;} //need to override it for POST request. 
	public abstract void setHttpHeaders(); //need to be override.
	public void streamContentOut(OutputStream outputStream) {}  //for voice search purpose.
	
	public void setTimeOut(int timeOut)
	{
		this.timeOut = timeOut;
	}

	public void setGZip(Boolean gzipped)
	{
		gzipUploadStream = gzipped;
	}

	protected void addHttpHeader(String name, String value)
	{
		if (name != null && name.length() > 0 && value != null && value.length() > 0)
			headers.put(name, value);
	}
	
	public void cancel()
	{
		if(future != null)
			future.cancel(true);
	}
		
	final public void run()  
	{		
		HttpURLConnection connection = null;		
		
		try {
			connection = (HttpURLConnection) new URL(url).openConnection();
			
			if (httpChunkedStreamingMode)
			{
				connection.setChunkedStreamingMode(m_ChunkedStreamSize);
			}
			
			connection.setConnectTimeout(timeOut);
			if (httpMethod == HttpMethod.HTTP_METHOD_GET)
				connection.setRequestMethod("GET");
			else if(httpMethod == HttpMethod.HTTP_METHOD_POST)
			{
				connection.setRequestMethod("POST");
				connection.setDoOutput(true);	
			}
			
			setHttpHeaders();
			
			Iterator<Entry<String, String>> headerIter = headers.entrySet().iterator(); 
			while(headerIter.hasNext())
			{
				Entry<String,String> item = headerIter.next();
				connection.addRequestProperty(item.getKey(),item.getValue());
			}
			
			connection.setDoInput(true);			
			connection.connect();
			
			//write content to outputstream
			if(httpMethod == HttpMethod.HTTP_METHOD_POST)
			{
				OutputStream outputStream = connection.getOutputStream();
				String msg = encode();
				if (msg != null && msg.length() > 0)
				{
					byte[] content = msg.getBytes("UTF-8");				
					
					if (outputStream != null)
					{
						if (gzipUploadStream)
						{
							ByteArrayOutputStream baos = new ByteArrayOutputStream();
							outputStream = new GZIPOutputStream(baos, 9);						
						}
						
						outputStream.write(content);						
					}
				}
				
				streamContentOut(outputStream);				
				outputStream.close();
			}
						
			int statusCode = connection.getResponseCode();			
			if (statusCode != HttpURLConnection.HTTP_OK) {
				String errorMsg = connection.getResponseMessage();
				httpResponse = new HttpResponseBase(HttpResponseBase.ERROR, statusCode, errorMsg);
			} else {
				Iterator<Entry<String, List<String>>> iter = connection.getHeaderFields().entrySet().iterator();
				Hashtable<String, String> responseHeader = new Hashtable<String, String> ();
				while(iter.hasNext())
				{
					Entry<String, List<String>> item = iter.next();
					String key = item.getKey();
					List<String> values = item.getValue();
					if (key != null && key.length() > 0 & !values.isEmpty())
					{
						String value = values.get(0);
						if (value != null && value.length() > 0)						
							responseHeader.put(key, value);						
					}
				}
				
				InputStream inputStream = connection.getInputStream();
                String encoding = connection.getContentEncoding();
                if (encoding != null){
                	encoding = encoding.toLowerCase();
                	if (encoding.indexOf("gzip") >= 0)
                		inputStream = new GZIPInputStream(inputStream);
                }
               
				httpResponse = decode(connection.getInputStream(), responseHeader);
				
			}				
			
		}catch (Exception e) {
			httpResponse = new HttpResponseBase(HttpResponseBase.ERROR, HttpResponseBase.NO_ERROR_CODE, e.toString());
		}
		finally
		{
			if (httpCallback != null)
				httpCallback.onResult(httpResponse);
		}
	}
}
